@@ -1,6 +1,9 @@
 mod ast;
+pub mod conjunction;
+pub mod disjunction;
 pub mod model;
 pub mod operators;
+
 use boolean_expression::Expr;
 lalrpop_mod!(
     #[allow(clippy::all)]
@@ -8,29 +11,36 @@ lalrpop_mod!(
     "/annis/db/aql/parser.rs"
 );
 
+use crate::annis::db::aql::conjunction::Conjunction;
+use crate::annis::db::aql::disjunction::Disjunction;
 use crate::annis::db::aql::operators::{
-    EqualValueSpec, IdenticalNodeSpec, PartOfSubCorpusSpec, RangeSpec,
+    EqualValueSpec, IdenticalNodeSpec, NegatedOpSpec, NonExistingUnaryOperatorSpec,
+    PartOfSubCorpusSpec, RangeSpec,
 };
 use crate::annis::db::exec::nodesearch::NodeSearchSpec;
-use crate::annis::db::query::conjunction::Conjunction;
-use crate::annis::db::query::disjunction::Disjunction;
 use crate::annis::errors::*;
 use crate::annis::operator::{BinaryOperatorSpec, UnaryOperatorSpec};
 use crate::annis::types::{LineColumn, LineColumnRange};
 use lalrpop_util::ParseError;
 use std::collections::BTreeMap;
 use std::collections::HashMap;
+use std::sync::Arc;
 
 thread_local! {
     static AQL_PARSER: parser::DisjunctionParser = parser::DisjunctionParser::new();
 }
 
-fn map_conjunction<'a>(
+#[derive(Clone, Default, Debug)]
+pub struct Config {
+    pub use_parallel_joins: bool,
+}
+
+fn map_conjunction(
     c: Vec<ast::Literal>,
     offsets: &BTreeMap<usize, usize>,
     var_idx_offset: usize,
     quirks_mode: bool,
-) -> Result<Conjunction<'a>> {
+) -> Result<Conjunction> {
     let mut q = Conjunction::with_offset(var_idx_offset);
     // collect and sort all node searches according to their start position in the text
     let (pos_to_node, pos_to_endpos) = calculate_node_positions(&c, offsets, quirks_mode)?;
@@ -46,14 +56,10 @@ fn map_conjunction<'a>(
                 ast::NodeRef::Name(name) => name.clone(),
             };
 
-            let op_pos: Option<LineColumnRange> = if let Some(pos) = pos {
-                Some(LineColumnRange {
-                    start: get_line_and_column_for_pos(pos.start, &offsets),
-                    end: Some(get_line_and_column_for_pos(pos.end, &offsets)),
-                })
-            } else {
-                None
-            };
+            let op_pos: Option<LineColumnRange> = pos.as_ref().map(|pos| LineColumnRange {
+                start: get_line_and_column_for_pos(pos.start, offsets),
+                end: Some(get_line_and_column_for_pos(pos.end, offsets)),
+            });
 
             q.add_unary_operator_from_query(make_unary_operator_spec(op.clone()), &var, op_pos)?;
         }
@@ -68,6 +74,7 @@ fn map_conjunction<'a>(
             mut op,
             rhs,
             pos,
+            negated,
         } = literal
         {
             let var_left = match lhs {
@@ -92,17 +99,13 @@ fn map_conjunction<'a>(
                 },
             };
 
-            let op_pos: Option<LineColumnRange> = if let Some(pos) = pos {
-                Some(LineColumnRange {
-                    start: get_line_and_column_for_pos(pos.start, &offsets),
-                    end: Some(get_line_and_column_for_pos(pos.end, &offsets)),
-                })
-            } else {
-                None
-            };
+            let op_pos: Option<LineColumnRange> = pos.map(|pos| LineColumnRange {
+                start: get_line_and_column_for_pos(pos.start, offsets),
+                end: Some(get_line_and_column_for_pos(pos.end, offsets)),
+            });
 
-            let spec_left = q.resolve_variable(&var_left, op_pos.clone())?;
-            let spec_right = q.resolve_variable(&var_right, op_pos.clone())?;
+            let node_left = q.resolve_variable(&var_left, op_pos.clone())?;
+            let node_right = q.resolve_variable(&var_right, op_pos.clone())?;
 
             if quirks_mode {
                 match op {
@@ -141,8 +144,51 @@ fn map_conjunction<'a>(
                     _ => {}
                 }
             }
-            let op_spec = make_binary_operator_spec(op, spec_left, spec_right)?;
-            q.add_operator_from_query(op_spec, &var_left, &var_right, op_pos, !quirks_mode)?;
+            let mut op_spec =
+                make_binary_operator_spec(op, node_left.spec.clone(), node_right.spec.clone())?;
+            if negated {
+                if !node_left.optional && !node_right.optional {
+                    op_spec = Arc::new(NegatedOpSpec {
+                        spec_left: node_left.spec.clone(),
+                        spec_right: node_right.spec.clone(),
+                        negated_op: op_spec,
+                    });
+                    q.add_operator_from_query(op_spec, &var_left, &var_right, op_pos, !quirks_mode)?
+                } else if node_left.optional && node_right.optional {
+                    // Not supported yet
+                    return Err(GraphAnnisError::AQLSemanticError(AQLError {
+                    desc: format!(
+                        "Negated binary operator needs a non-optional left or right operand, but both operands (#{}, #{}) are optional, as indicated by their \"?\" suffix.", 
+                        var_left, var_right),
+                    location: op_pos,
+                }));
+                } else {
+                    let target_left = node_left.optional;
+                    let filtered_var = if target_left {
+                        node_right.var
+                    } else {
+                        node_left.var
+                    };
+                    let spec = NonExistingUnaryOperatorSpec {
+                        op: op_spec,
+                        target: if target_left {
+                            node_left.spec
+                        } else {
+                            node_right.spec
+                        },
+                        target_left,
+                    };
+                    q.add_unary_operator_from_query(Arc::new(spec), &filtered_var, op_pos)?;
+                }
+            } else if node_left.optional || node_right.optional {
+                // Not supported yet
+                return Err(GraphAnnisError::AQLSemanticError(AQLError {
+                    desc: "Optional left or right operands can only be combined with a negated operator.".into(),
+                    location: op_pos,
+                }));
+            } else {
+                q.add_operator_from_query(op_spec, &var_left, &var_right, op_pos, !quirks_mode)?;
+            }
         }
     }
 
@@ -154,9 +200,9 @@ fn map_conjunction<'a>(
             let num_joins = num_pointing_or_dominance_joins.get(orig_var).unwrap_or(&0);
             // add an additional node for each extra join and join this artificial node with identity relation
             for _ in 1..*num_joins {
-                if let Ok(node_spec) = q.resolve_variable(&orig_var, None) {
-                    let new_var = q.add_node(node_spec, None);
-                    q.add_operator(Box::new(IdenticalNodeSpec {}), &orig_var, &new_var, false)?;
+                if let Ok(node) = q.resolve_variable(orig_var, None) {
+                    let new_var = q.add_node(node.spec, None);
+                    q.add_operator(Arc::new(IdenticalNodeSpec {}), orig_var, &new_var, false)?;
                 }
             }
         }
@@ -165,7 +211,7 @@ fn map_conjunction<'a>(
     Ok(q)
 }
 
-type PosToNodeMap = BTreeMap<usize, (NodeSearchSpec, Option<String>)>;
+type PosToNodeMap = BTreeMap<usize, (NodeSearchSpec, Option<String>, bool)>;
 type PosToEndPosMap = BTreeMap<usize, usize>;
 
 fn calculate_node_positions(
@@ -182,9 +228,10 @@ fn calculate_node_positions(
                 spec,
                 pos,
                 variable,
+                optional,
             } => {
                 if let Some(pos) = pos {
-                    pos_to_node.insert(pos.start, (spec.clone(), variable.clone()));
+                    pos_to_node.insert(pos.start, (spec.clone(), variable.clone(), *optional));
                     pos_to_endpos.insert(pos.start, pos.end);
                 }
             }
@@ -193,22 +240,24 @@ fn calculate_node_positions(
                     spec,
                     pos,
                     variable,
+                    optional,
                 } = lhs
                 {
                     pos_to_node
                         .entry(pos.start)
-                        .or_insert_with(|| (spec.as_ref().clone(), variable.clone()));
+                        .or_insert_with(|| (spec.as_ref().clone(), variable.clone(), *optional));
                     pos_to_endpos.entry(pos.start).or_insert_with(|| pos.end);
                 }
                 if let ast::Operand::Literal {
                     spec,
                     pos,
                     variable,
+                    optional,
                 } = rhs
                 {
                     pos_to_node
                         .entry(pos.start)
-                        .or_insert_with(|| (spec.as_ref().clone(), variable.clone()));
+                        .or_insert_with(|| (spec.as_ref().clone(), variable.clone(), *optional));
                     pos_to_endpos.entry(pos.start).or_insert_with(|| pos.end);
                 }
             }
@@ -217,10 +266,10 @@ fn calculate_node_positions(
             }
             ast::Literal::LegacyMetaSearch { pos, .. } => {
                 if !quirks_mode {
-                    let start = get_line_and_column_for_pos(pos.start, &offsets);
+                    let start = get_line_and_column_for_pos(pos.start, offsets);
                     let end = Some(get_line_and_column_for_pos(
                         pos.start + "meta::".len() - 1,
-                        &offsets,
+                        offsets,
                     ));
                     return Err(GraphAnnisError::AQLSyntaxError( AQLError {
                         desc: "Legacy metadata search is no longer allowed. Use the @* operator and normal attribute search instead.".into(),
@@ -234,26 +283,25 @@ fn calculate_node_positions(
     Ok((pos_to_node, pos_to_endpos))
 }
 
-fn add_node_specs_by_start<'a>(
-    q: &mut Conjunction<'a>,
-    pos_to_node: BTreeMap<usize, (NodeSearchSpec, Option<String>)>,
+fn add_node_specs_by_start(
+    q: &mut Conjunction,
+    pos_to_node: BTreeMap<usize, (NodeSearchSpec, Option<String>, bool)>,
     pos_to_endpos: BTreeMap<usize, usize>,
     offsets: &BTreeMap<usize, usize>,
 ) -> Result<BTreeMap<usize, String>> {
     let mut pos_to_node_id: BTreeMap<usize, String> = BTreeMap::default();
-    for (start_pos, (node_spec, variable)) in pos_to_node {
-        let start = get_line_and_column_for_pos(start_pos, &offsets);
-        let end = if let Some(end_pos) = pos_to_endpos.get(&start_pos) {
-            Some(get_line_and_column_for_pos(*end_pos, &offsets))
-        } else {
-            None
-        };
+    for (start_pos, (node_spec, variable, optional)) in pos_to_node {
+        let start = get_line_and_column_for_pos(start_pos, offsets);
+        let end = pos_to_endpos
+            .get(&start_pos)
+            .map(|end_pos| get_line_and_column_for_pos(*end_pos, offsets));
 
         let idx = q.add_node_from_query(
             node_spec,
             variable.as_deref(),
             Some(LineColumnRange { start, end }),
             true,
+            optional,
         );
         pos_to_node_id.insert(start_pos, idx.clone());
     }
@@ -271,11 +319,11 @@ fn add_legacy_metadata_constraints(
         // TODO: add warning to the user not to use this construct anymore
         for (spec, _pos) in legacy_meta_search {
             // add an artificial node that describes the document/corpus node
-            let meta_node_idx = q.add_node_from_query(spec, None, None, false);
+            let meta_node_idx = q.add_node_from_query(spec, None, None, false, false);
             if let Some(first_meta_idx) = first_meta_idx.clone() {
                 // avoid nested loops by joining additional meta nodes with a "identical node"
                 q.add_operator(
-                    Box::new(IdenticalNodeSpec {}),
+                    Arc::new(IdenticalNodeSpec {}),
                     &first_meta_idx,
                     &meta_node_idx,
                     true,
@@ -284,7 +332,7 @@ fn add_legacy_metadata_constraints(
                 first_meta_idx = Some(meta_node_idx.clone());
                 // add a special join to the first node of the query
                 q.add_operator(
-                    Box::new(PartOfSubCorpusSpec {
+                    Arc::new(PartOfSubCorpusSpec {
                         dist: RangeSpec::Unbound,
                     }),
                     &first_node_pos,
@@ -303,9 +351,10 @@ fn add_legacy_metadata_constraints(
                     None,
                     None,
                     false,
+                    false,
                 );
                 q.add_operator(
-                    Box::new(IdenticalNodeSpec {}),
+                    Arc::new(IdenticalNodeSpec {}),
                     &meta_node_idx,
                     &doc_anno_idx,
                     true,
@@ -368,7 +417,7 @@ fn get_alternatives_from_dnf(expr: ast::Expr) -> Vec<Vec<ast::Literal>> {
     vec![]
 }
 
-pub fn parse<'a>(query_as_aql: &str, quirks_mode: bool) -> Result<Disjunction<'a>> {
+pub fn parse(query_as_aql: &str, quirks_mode: bool) -> Result<Disjunction> {
     let ast = AQL_PARSER.with(|p| p.parse(query_as_aql));
     match ast {
         Ok(ast) => {
@@ -437,26 +486,26 @@ fn make_binary_operator_spec(
     op: ast::BinaryOpSpec,
     spec_left: NodeSearchSpec,
     spec_right: NodeSearchSpec,
-) -> Result<Box<dyn BinaryOperatorSpec>> {
-    let op_spec: Box<dyn BinaryOperatorSpec> = match op {
-        ast::BinaryOpSpec::Dominance(spec) => Box::new(spec),
-        ast::BinaryOpSpec::Pointing(spec) => Box::new(spec),
-        ast::BinaryOpSpec::Precedence(spec) => Box::new(spec),
-        ast::BinaryOpSpec::Near(spec) => Box::new(spec),
-        ast::BinaryOpSpec::Overlap(spec) => Box::new(spec),
-        ast::BinaryOpSpec::IdenticalCoverage(spec) => Box::new(spec),
-        ast::BinaryOpSpec::PartOfSubCorpus(spec) => Box::new(spec),
-        ast::BinaryOpSpec::Inclusion(spec) => Box::new(spec),
-        ast::BinaryOpSpec::LeftAlignment(spec) => Box::new(spec),
-        ast::BinaryOpSpec::RightAlignment(spec) => Box::new(spec),
-        ast::BinaryOpSpec::IdenticalNode(spec) => Box::new(spec),
+) -> Result<Arc<dyn BinaryOperatorSpec>> {
+    let op_spec: Arc<dyn BinaryOperatorSpec> = match op {
+        ast::BinaryOpSpec::Dominance(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::Pointing(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::Precedence(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::Near(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::Overlap(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::IdenticalCoverage(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::PartOfSubCorpus(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::Inclusion(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::LeftAlignment(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::RightAlignment(spec) => Arc::new(spec),
+        ast::BinaryOpSpec::IdenticalNode(spec) => Arc::new(spec),
         ast::BinaryOpSpec::ValueComparison(cmp) => match cmp {
-            ast::ComparisonOperator::Equal => Box::new(EqualValueSpec {
+            ast::ComparisonOperator::Equal => Arc::new(EqualValueSpec {
                 spec_left,
                 spec_right,
                 negated: false,
             }),
-            ast::ComparisonOperator::NotEqual => Box::new(EqualValueSpec {
+            ast::ComparisonOperator::NotEqual => Arc::new(EqualValueSpec {
                 spec_left,
                 spec_right,
                 negated: true,
@@ -466,9 +515,9 @@ fn make_binary_operator_spec(
     Ok(op_spec)
 }
 
-fn make_unary_operator_spec(op: ast::UnaryOpSpec) -> Box<dyn UnaryOperatorSpec> {
+fn make_unary_operator_spec(op: ast::UnaryOpSpec) -> Arc<dyn UnaryOperatorSpec> {
     match op {
-        ast::UnaryOpSpec::Arity(spec) => Box::new(spec),
+        ast::UnaryOpSpec::Arity(spec) => Arc::new(spec),
     }
 }
 
